from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

from .auth_objects import export_codenames
from .constants import DATA_EXPORTER_ROLE, EXPORT, EXPORT_PII

site_auths.add_post_update_func(
    "edc_export",
    remove_default_model_permissions_from_edc_permissions,
)

site_auths.add_custom_permissions_tuples(
    model="edc_export.edcpermissions",
    codename_tuples=[
        (
            "edc_export.export_visitschedule",
            "Can export the visit schedule",
        ),
        (
            "edc_export.view_export_dashboard",
            "Can view export dashboard",
        ),
        (
            "edc_export.export_subjectschedulehistory",
            "Can export subject schedule history",
        ),
        (
            "edc_export.export_pii",
            "Can export PII",
        ),
    ],
)


site_auths.add_group(*export_codenames, name=EXPORT)
site_auths.add_role(EXPORT, name=DATA_EXPORTER_ROLE)
site_auths.add_group("edc_export.export_pii", name=EXPORT_PII)



from __future__ import annotations

from tempfile import mkdtemp
from typing import TYPE_CHECKING, Type

from edc_pdutils.df_exporters import CsvModelExporter
from edc_sites.site import sites
from edc_utils import get_utcnow

from .files_archiver import FilesArchiver
from .files_emailer import FilesEmailer, FilesEmailerError

if TYPE_CHECKING:
    from datetime import datetime

    from django.contrib.auth.models import User


class ArchiveExporterNothingExported(Exception):
    pass


class ArchiveExporterEmailError(Exception):
    pass


class ArchiveExporter:
    """Exports a list of models to individual CSV files and
    adds each to a single zip archive OR emails each.

    models: a list of model names in label_lower format.
    """

    date_format: str = "%Y%m%d%H%M%S"
    csv_exporter_cls: Type[CsvModelExporter] = CsvModelExporter
    files_emailer_cls: Type[FilesEmailer] = FilesEmailer
    files_archiver_cls: Type[FilesArchiver] = FilesArchiver

    def __init__(
        self,
        models: list[str] = None,
        decrypt: bool | None = None,
        user: User = None,
        archive: bool | None = None,
        email_to_user: bool | None = None,
        **kwargs,
    ):
        models: list[str] = models or []
        self.archive_filename: str | None = None
        self.exported: list = []
        self.emailed_to: str | None = None
        self.emailed_datetime: datetime | None = None
        self.exported_datetime: datetime | None = None
        tmp_folder: str = mkdtemp()
        for model in models:
            csv_exporter = self.csv_exporter_cls(
                model=model,
                export_folder=tmp_folder,
                decrypt=decrypt,
                site_ids=sites.get_site_ids_for_user(
                    user=user, site_id=sites.get_current_site().site_id
                ),
                **kwargs,
            )
            self.exported.append(csv_exporter.to_csv())
        if not self.exported:
            raise ArchiveExporterNothingExported(
                f"Nothing exported. Got models={models}."
            )
        else:
            if archive:
                archiver = self.files_archiver_cls(
                    path=tmp_folder,
                    user=user,
                    exported_datetime=self.exported_datetime,
                    date_format=self.date_format,
                )
                self.archive_filename = archiver.archive_filename
                self.exported_datetime = archiver.exported_datetime
            if email_to_user:
                summary = [str(x) for x in self.exported]
                summary.sort()
                try:
                    self.files_emailer_cls(
                        path=tmp_folder,
                        user=user,
                        file_ext=".zip" if archive else ".csv",
                        summary="\n".join(summary),
                    )
                except FilesEmailerError as e:
                    raise ArchiveExporterEmailError(e)
                else:
                    self.emailed_to = user.email
                    self.emailed_datetime = get_utcnow()
                    self.exported_datetime = (
                        self.exported_datetime or self.emailed_datetime
                    )



from edc_reportable.exceptions import NotEvaluated
from edc_reportable.utils import get_reference_range_collection

__all__ = ["get_summary"]


def get_summary(obj) -> tuple[list[str], list[str], list[str]]:
    """Returns a list of graded or abnormal values given a
    BloodResultsModel instance.
    """
    opts = obj.get_summary_options()
    reportable = []
    abnormal = []
    errors = []
    for field_name in [f.name for f in obj._meta.get_fields()]:
        try:
            utest_id, _ = field_name.split("_value")
        except ValueError:
            utest_id = field_name
        reference_range_collection = get_reference_range_collection(obj)
        if units := getattr(obj, f"{utest_id}_units", None):
            if value := getattr(obj, field_name):
                opts.update(units=units, label=utest_id)
                try:
                    grading_data, grading_eval_phrase = (
                        reference_range_collection.get_grade(value, **opts)
                    )
                except NotEvaluated as e:
                    errors.append(f"{e}.")
                    grading_data = None
                    grading_eval_phrase = None
                if (
                    grading_data
                    and grading_data.grade
                    in reference_range_collection.reportable_grades(utest_id)
                ):
                    setattr(obj, f"{utest_id}_grade", grading_data.grade)
                    setattr(
                        obj, f"{utest_id}_grade_description", grading_data.description
                    )
                    reportable.append(f"{grading_eval_phrase}")
                else:
                    try:
                        is_normal, normal_data = reference_range_collection.is_normal(
                            value, **opts
                        )
                    except NotEvaluated as e:
                        errors.append(f"{e}.")
                    else:
                        if is_normal is False:
                            abnormal.append(
                                f"{normal_data.label}: {value} {units} "
                                f"{normal_data.phrase} Abnormal"
                            )
    return reportable, abnormal, errors



from django.apps import AppConfig as DjangoAppConfig


class AppConfig(DjangoAppConfig):
    name = "edc_document_status"
    verbose_name = "Edc Document Status"
    has_exportable_data = False
    include_in_administration_section = False



from __future__ import annotations

from datetime import date, datetime
from zoneinfo import ZoneInfo

from django.conf import settings


class EdcDatetimeError(Exception):
    pass


def get_utcnow() -> datetime:
    return datetime.now().astimezone(ZoneInfo("UTC"))


def get_utcnow_as_date() -> date:
    return datetime.now().astimezone(ZoneInfo("UTC")).date()


def to_utc(dt: datetime) -> datetime:
    """Returns UTC datetime from any aware datetime."""
    return dt.astimezone(ZoneInfo("UTC"))


def to_local(dt: datetime) -> datetime:
    """Returns local datetime from any aware datetime."""
    return dt.astimezone(ZoneInfo(settings.TIME_ZONE))


def floor_secs(dte) -> datetime:
    return datetime(
        dte.year, dte.month, dte.day, dte.hour, dte.minute, 0, 0, tzinfo=dte.tzinfo
    )


def ceil_secs(dte) -> datetime:
    return datetime(
        dte.year,
        dte.month,
        dte.day,
        dte.hour,
        dte.minute,
        59,
        999999,
        tzinfo=dte.tzinfo,
    )


def floor_datetime(dt) -> datetime:
    return dt.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=dt.tzinfo)


def ceil_datetime(dt) -> datetime:
    return dt.replace(
        hour=23, minute=59, second=59, microsecond=999999, tzinfo=dt.tzinfo
    )



import csv

from django.apps import apps as django_apps

from edc_utils import get_utcnow


class FileHistoryUpdater:
    file_history_model = "edc_export.filehistory"

    def __init__(
        self,
        model=None,
        filename=None,
        notification_plan_name=None,
        path=None,
        delimiter=None,
    ):
        self.model = model
        self.filename = filename
        self.notification_plan_name = notification_plan_name
        self.path = path
        self.delimiter = delimiter

    @property
    def model_cls(self):
        return django_apps.get_model(self.file_history_model)

    def update(self):
        exported_pks = []
        export_uuids = []
        with open(self.path, "r") as f:
            csv_reader = csv.DictReader(f, delimiter=self.delimiter)
            for row in csv_reader:
                exported_pks.append(row["id"])
                export_uuids.append(row["export_uuid"])
        return self.model_cls.objects.create(
            model=self.model,
            pk_list="|".join(exported_pks),
            export_uuid_list="|".join(export_uuids),
            exported=True,
            exported_datetime=get_utcnow(),
            filename=self.filename,
            notification_plan_name=self.notification_plan_name,
        )



from typing import Optional

from .model_notification import ModelNotification


class GradedEventNotification(ModelNotification):
    grade: Optional[int] = None
    model: Optional[str] = None
    create_fields = ["ae_grade"]
    update_fields = ["ae_grade"]

    def field_value_condition_on_create(self, field, current_value):
        return str(current_value) == str(self.grade)

    def field_value_condition_on_update(self, field, previous_value, current_value):
        """Returns True if the value has changed and matches self.grade"""
        return str(previous_value) != str(current_value) and str(current_value) == str(
            self.grade
        )



from edc_auth.site_auths import site_auths

from .auth_objects import EDC_FACILITY, EDC_FACILITY_SUPER, EDC_FACILITY_VIEW, codenames

site_auths.add_group(*codenames, name=EDC_FACILITY_VIEW, view_only=True)
site_auths.add_group(*codenames, name=EDC_FACILITY, no_delete=True)
site_auths.add_group(*codenames, name=EDC_FACILITY_SUPER)



# Generated by Django 3.2.13 on 2022-09-29 14:42

from django.db import migrations
import django.db.models.manager
import edc_sites.models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0023_auto_20220704_1841"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="manifestitem",
            managers=[
                ("objects", django.db.models.manager.Manager()),
                ("on_site", edc_sites.models.CurrentSiteManager()),
            ],
        ),
    ]



from django.contrib.sites.models import Site
from django.db import models

from edc_utils import get_utcnow

from ..model_mixins import qa_reports_permissions


class QaReportLog(models.Model):

    username = models.CharField(max_length=100)
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    report_model = models.CharField(max_length=100)
    accessed = models.DateTimeField(default=get_utcnow)

    class Meta:
        verbose_name = "QA Report Log"
        verbose_name_plural = "QA Report Logs"
        indexes = [models.Index(fields=["accessed"])]
        default_permissions = qa_reports_permissions



from django.contrib.sites.models import Site
from django.db import models
from django.db.models import DO_NOTHING, Index

from edc_utils import get_utcnow

from ..utils import recreate_db_view
from .qa_reports_permissions import qa_reports_permissions


class QaReportModelMixin(models.Model):

    report_model = models.CharField(max_length=50)

    subject_identifier = models.CharField(max_length=25)

    site = models.ForeignKey(Site, on_delete=DO_NOTHING)

    created = models.DateTimeField(default=get_utcnow)

    @classmethod
    def recreate_db_view(cls, drop: bool | None = None, verbose: bool | None = None):
        recreate_db_view(cls, drop=drop, verbose=verbose)

    class Meta:
        abstract = True
        default_permissions = qa_reports_permissions
        indexes = [Index(fields=["subject_identifier", "site"])]



from datetime import date, datetime, timedelta

from django.core.exceptions import ValidationError

from edc_utils import get_utcnow


def datetime_not_future(utc_datetime: datetime) -> None:
    time_error = timedelta(minutes=10)
    if utc_datetime > get_utcnow() + time_error:
        raise ValidationError("Cannot be a future date/time")


def date_not_future(value: date) -> None:
    if value > get_utcnow().date():
        raise ValidationError("Cannot be a future date")


def date_is_past(value: date) -> None:
    if value > get_utcnow().date():
        raise ValidationError("Expected a past date")


def date_is_not_now(value: date) -> None:
    if value == get_utcnow().date():
        raise ValidationError("Cannot be today")


def datetime_is_future(utc_datetime: datetime) -> None:
    time_error = timedelta(minutes=10)
    if utc_datetime < get_utcnow() + time_error:
        raise ValidationError("Expected a future date/time")


def date_is_future(value: date) -> None:
    if value < get_utcnow().date():
        raise ValidationError("Expected a future date")



from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions
from edc_lab.auth_objects import LAB, LAB_VIEW
from edc_lab_dashboard.auth_objects import (
    lab_dashboard_codenames,
    lab_dashboard_tuples,
    lab_navbar_codenames,
    lab_navbar_tuples,
)

site_auths.add_post_update_func(
    "edc_lab_dashboard", remove_default_model_permissions_from_edc_permissions
)

site_auths.add_custom_permissions_tuples(
    model="edc_lab_dashboard.edcpermissions", codename_tuples=lab_navbar_tuples
)
site_auths.add_custom_permissions_tuples(
    model="edc_lab_dashboard.edcpermissions", codename_tuples=lab_dashboard_tuples
)

site_auths.update_group(*lab_dashboard_codenames, *lab_navbar_codenames, name=LAB)
site_auths.update_group(*lab_dashboard_codenames, *lab_navbar_codenames, name=LAB_VIEW)



from django.db import models
from django.utils import timezone

from edc_constants.constants import CLOSED, OPEN, OTHER
from edc_sites.model_mixins import SiteModelMixin
from edc_utils import get_utcnow

from ...constants import STORAGE, TESTING
from ...identifiers import ManifestIdentifier

STATUS = ((OPEN, "Open"), (CLOSED, "Closed"))

MANIFEST_CATEGORY = ((TESTING, "Testing"), (STORAGE, "Storage"), (OTHER, "Other"))


class ManifestModelMixin(SiteModelMixin, models.Model):
    manifest_identifier = models.CharField(
        verbose_name="Manifest Identifier", max_length=25, editable=False, unique=True
    )

    manifest_datetime = models.DateTimeField(default=timezone.now)

    export_datetime = models.DateTimeField(null=True, blank=True)

    export_references = models.TextField(null=True, blank=True)

    description = models.TextField(
        verbose_name="Description of contents",
        null=True,
        help_text="If blank will be automatically generated",
    )

    status = models.CharField(max_length=15, default=OPEN, choices=STATUS)

    category = models.CharField(
        max_length=25, default=TESTING, choices=MANIFEST_CATEGORY
    )

    category_other = models.CharField(max_length=25, null=True, blank=True)

    comment = models.TextField(verbose_name="Comment", null=True)

    shipped = models.BooleanField(default=False)

    printed = models.BooleanField(default=False)

    printed_datetime = models.DateTimeField(null=True, blank=True)

    def save(self, *args, **kwargs):
        if not self.manifest_identifier:
            identifier = ManifestIdentifier()
            self.manifest_identifier = identifier.identifier
        if self.shipped and not self.export_datetime:
            self.export_datetime = get_utcnow()
        elif not self.shipped:
            self.export_datetime = None
            self.printed = False
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.manifest_identifier,)  # noqa

    natural_key.dependencies = ["sites.Site"]

    @property
    def human_readable_identifier(self):
        x = self.manifest_identifier
        return "{}-{}-{}".format(x[0:4], x[4:8], x[8:12])

    class Meta:
        abstract = True



from edc_adverse_event.constants import TMG_ROLE
from edc_auth.constants import (
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    CLINICIAN_SUPER_ROLE,
    NURSE_ROLE,
)
from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions

REVIEW = "REVIEW"

site_auths.add_post_update_func(
    "edc_review_dashboard", remove_default_model_permissions_from_edc_permissions
)

site_auths.add_custom_permissions_tuples(
    model="edc_review_dashboard.edcpermissions",
    codename_tuples=[
        (
            "edc_review_dashboard.view_subject_review_listboard",
            "Can access subject review listboard",
        )
    ],
)
site_auths.add_group("edc_review_dashboard.view_subject_review_listboard", name=REVIEW)

site_auths.update_role(REVIEW, name=AUDITOR_ROLE)
site_auths.update_role(REVIEW, name=TMG_ROLE)
site_auths.update_role(REVIEW, name=CLINICIAN_ROLE)
site_auths.update_role(REVIEW, name=CLINICIAN_SUPER_ROLE)
site_auths.update_role(REVIEW, name=NURSE_ROLE)



# default group names

ACCOUNT_MANAGER = "ACCOUNT_MANAGER"
ADMINISTRATION = "ADMINISTRATION"
AUDITOR = "AUDITOR"
CELERY_MANAGER = "CELERY_MANAGER"
CLINIC = "CLINIC"
CLINIC_SUPER = "CLINIC_SUPER"
EVERYONE = "EVERYONE"
PII = "PII"
PII_VIEW = "PII_VIEW"

# default role names
ACCOUNT_MANAGER_ROLE = "account_manager"
AUDITOR_ROLE = "auditor"
CLINICIAN_ROLE = "clinician"
CLINICIAN_SUPER_ROLE = "clinician_super"
CUSTOM_ROLE = "custom"
NURSE_ROLE = "research_nurse"
STAFF_ROLE = "staff"
STATISTICIAN = "statistician"
